@@ -35,6 +35,7 @@
 #include "config.h"
 #endif
 
+#include <assert.h>
 #include <errno.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -131,6 +132,7 @@
 #endif
 
 #include "arping.h"
+#include "cast.h"
 
 #ifndef ETH_ALEN
 #define ETH_ALEN 6
@@ -150,6 +152,10 @@
 
 #define UNUSED(x) (void)(x)
 
+#ifndef PCAP_NETMASK_UNKNOWN
+#define PCAP_NETMASK_UNKNOWN 0xffffffff
+#endif
+
 /**
  * OS-specific interface finding using routing table. See findif_*.c
  * ebuf must be called with a size of at least
@@ -186,7 +192,7 @@ uint32_t dstip;
 static uint8_t dstmac[ETH_ALEN];
 
 static char* payload_suffix = NULL;
-static ssize_t payload_suffix_size = -1;
+static size_t payload_suffix_size = 4;
 
 uint32_t srcip;                   /* autodetected, override with -S/-b/-0 */
 uint8_t srcmac[ETH_ALEN];         /* autodetected, override with -s */
@@ -238,17 +244,156 @@ int verbose = 0;  /* Increase with -v */
 /* Doesn't really need to be volatile, but doesn't hurt. */
 static volatile sig_atomic_t time_to_die = 0;
 
+static float
+must_parse_float(const char* in, const char* what)
+{
+        if (!*in) {
+                fprintf(stderr, "arping: %s: value was empty\n", what);
+                exit(1);
+        }
+        char *endp = NULL;
+        errno = 0;
+        // Maybe need to use strtod() instead? It's in C89, whereas
+        // looks like strtof() is in POSIX.1-2001, POSIX.1-2008, C99.
+        const float ret = strtof(in, &endp);
+        if (errno) {
+                fprintf(stderr, "arping: %s: parsing <%s> as float: %s\n",
+                        what, in, strerror(errno));
+                exit(1);
+        }
+        if (*endp) {
+                fprintf(stderr,
+                        "arping: %s: failed parsing <%s> as float\n",
+                        what, in);
+                exit(1);
+        }
+        return ret;
+}
+
+static int
+must_parse_int(const char* in, const char* what)
+{
+        if (!*in) {
+                fprintf(stderr, "arping: %s: value was empty\n", what);
+                exit(1);
+        }
+        char *endp = NULL;
+        errno = 0;
+        const long ret = strtol(in, &endp, 0);
+        if (errno) {
+                fprintf(stderr, "arping: %s: parsing <%s> as integer: %s\n",
+                        what, in, strerror(errno));
+                exit(1);
+        }
+        if (*endp) {
+                fprintf(stderr,
+                        "arping: %s: failed parsing <%s> as integer\n",
+                        what, in);
+                exit(1);
+        }
+        return cast_long_int(ret, "%s %s", what, in);
+}
+
+static int16_t
+must_parse_int16(const char* in, const char* what)
+{
+        const int out = must_parse_int(in, what);
+        return cast_int_int16(out, "%s %s", what, in);
+}
+
+/*
+ * It's not actually possible to parse unsigned numbers in C.
+ * The best this function can do is parse as a long, then reject
+ * negative numbers. If sizeof(int) == sizeof(long) then that cuts
+ * off half of all possible values.
+ *
+ * More ranting here:
+ * https://blog.habets.se/2022/10/No-way-to-parse-integers-in-C.html
+ */
+static unsigned int
+must_parse_uint(const char* in, const char* what)
+{
+        if (!*in) {
+                fprintf(stderr, "arping: %s: value was empty\n", what);
+                exit(1);
+        }
+        char *endp = NULL;
+        errno = 0;
+#if HAVE_STRTOLL
+#define CASTER cast_longlong_uint
+        const long long ret = strtoll(in, &endp, 0);
+#else
+#define CASTER cast_long_uint
+        const long ret = strtol(in, &endp, 0);
+#endif
+        if (errno) {
+                fprintf(stderr, "arping: %s: parsing <%s> as integer: %s\n",
+                        what, in, strerror(errno));
+                exit(1);
+        }
+        if (*endp) {
+                fprintf(stderr,
+                        "arping: %s: failed parsing <%s> as integer\n",
+                        what, in);
+                exit(1);
+        }
+        if (ret < 0) {
+                fprintf(stderr,
+                        "arping: %s: <%s> is negative\n", what, in);
+                exit(1);
+        }
+        return CASTER(ret, "%s %s", what, in);
+#undef CASTER
+}
+
+static unsigned long
+parse_ulong(const char* in, const char* what, char* ebuf, size_t ebuflen)
+{
+        if (!*in) {
+                snprintf(ebuf, ebuflen, "arping: %s: value was empty\n", what);
+                exit(1);
+        }
+        char *endp = NULL;
+        errno = 0;
+#if HAVE_STRTOLL
+#define CASTER cast_longlong_ulonglong
+        const long long ret = strtoll(in, &endp, 0);
+#else
+#define CASTER cast_long_ulong
+        const long ret = strtol(in, &endp, 0);
+#endif
+        if (errno) {
+                snprintf(ebuf, ebuflen, "arping: %s: parsing <%s> as integer: %s\n",
+                        what, in, strerror(errno));
+                exit(1);
+        }
+        if (*endp) {
+                snprintf(ebuf, ebuflen,
+                        "arping: %s: failed parsing <%s> as integer\n",
+                        what, in);
+                exit(1);
+        }
+        if (ret < 0) {
+                snprintf(ebuf, ebuflen,
+                        "arping: %s: <%s> is negative\n", what, in);
+                exit(1);
+        }
+        return CASTER(ret, "%s %s", what, in);
+#undef CASTER
+}
+
 static ssize_t
 xgetrandom(void *buf, const size_t buflen, const unsigned int flags)
 {
 #ifdef HAVE_GETRANDOM
         return getrandom(buf, buflen, flags);
 #else
-        char* p = buf;
-        for (int n = 0; n < buflen; n++) {
-                p[n] = random() & 0xff;
+        UNUSED(flags);
+        uint8_t* p = buf;
+        for (size_t n = 0; n < buflen; n++) {
+                p[n] = cast_long_uint8(random() & 0xff, NULL);
         }
-        return buflen;
+        return cast_size_ssize(buflen, NULL);
 #endif
 }
 
@@ -398,10 +543,10 @@ must_get_group(const char* ident)
 
                 // Not a name. Try it as an integer.
                 {
-                        char* endp = NULL;
-                        gid_t r = strtol(ident, &endp, 0);
-                        if (!*endp) {
-                                return r;
+                        char ebuf[1024] = {0};
+                        const unsigned long v = parse_ulong(ident, "gid", ebuf, sizeof(ebuf));
+                        if (!ebuf[0]) {
+                                return cast_ulong_gid(v, NULL);
                         }
                 }
         }
@@ -514,11 +659,11 @@ static void drop_seccomp(int libnet_fd)
         }
 
         // Libnet.
-        if (seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(ioctl), 1, SCMP_A0(SCMP_CMP_EQ, libnet_fd))) {
+        if (seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(ioctl), 1, SCMP_A0(SCMP_CMP_EQ, cast_int_uint(libnet_fd, NULL)))) {
                 perror("seccomp_rule_add(ioctl libnet)");
                 exit(1);
         }
-        if (seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(sendto), 1, SCMP_A0(SCMP_CMP_EQ, libnet_fd))) {
+        if (seccomp_rule_add(ctx, SCMP_ACT_ALLOW, SCMP_SYS(sendto), 1, SCMP_A0(SCMP_CMP_EQ, cast_int_uint(libnet_fd, NULL)))) {
                 perror("seccomp_rule_add(sendto libnet)");
                 exit(1);
         }
@@ -550,6 +695,7 @@ static void drop_seccomp(int libnet_fd)
 static void
 drop_more_privileges(int libnet_fd)
 {
+        UNUSED(libnet_fd);
 #ifdef HAVE_PLEDGE
         if (pledge("stdio tty", "")) {
                 fprintf(stderr, "arping: failed to pledge(stdio, <empty>): %s\n",
@@ -609,8 +755,7 @@ bug_pcap_vlan()
  * FIXME: Use pcap_set_buffer_size()?
  */
 static pcap_t*
-try_pcap_open_live(const char *device, int snaplen,
-                   int promisc, int to_ms, char *errbuf)
+try_pcap_open_live(const char *device, int snaplen, int to_ms, char *errbuf)
 {
 #ifdef HAVE_PCAP_CREATE
         pcap_t* pcap;
@@ -694,7 +839,7 @@ err:
         }
         return NULL;
 #else
-        return pcap_open_live(device, snaplen, promisc, to_ms, errbuf);
+        return pcap_open_live(device, snaplen, to_ms, errbuf);
 #endif
 }
 
@@ -709,28 +854,27 @@ err:
  * fails to open /dev/net, because it's a directory.
  */
 static pcap_t*
-do_pcap_open_live(const char *device, int snaplen,
-                  int promisc, int to_ms, char *errbuf)
+do_pcap_open_live(const char *device, int snaplen, int to_ms, char *errbuf)
 {
         pcap_t* ret;
         char buf[PATH_MAX];
 
-        if ((ret = try_pcap_open_live(device, snaplen, promisc, to_ms, errbuf))) {
+        if ((ret = try_pcap_open_live(device, snaplen, to_ms, errbuf))) {
                 return ret;
         }
 
         snprintf(buf, sizeof(buf), "/dev/%s", device);
-        if ((ret = try_pcap_open_live(buf, snaplen, promisc, to_ms, errbuf))) {
+        if ((ret = try_pcap_open_live(buf, snaplen, to_ms, errbuf))) {
                 return ret;
         }
 
         snprintf(buf, sizeof(buf), "/dev/net/%s", device);
-        if ((ret = try_pcap_open_live(buf, snaplen, promisc, to_ms, errbuf))) {
+        if ((ret = try_pcap_open_live(buf, snaplen, to_ms, errbuf))) {
                 return ret;
         }
 
         /* Call original again to reset the error message. */
-        return try_pcap_open_live(device, snaplen, promisc, to_ms, errbuf);
+        return try_pcap_open_live(device, snaplen, to_ms, errbuf);
 }
 
 /**
@@ -1107,7 +1251,7 @@ update_stats(double sample)
 static double
 timespec2dbl(const struct timespec *tv)
 {
-        return tv->tv_sec + (double)tv->tv_nsec / 1000000000;
+        return (double)tv->tv_sec + (double)tv->tv_nsec / 1000000000;
 }
 
 /**
@@ -1131,7 +1275,10 @@ wait_time(double deadline, uint32_t packetwait)
         if (max_wait > packetwait / 1000000.0) {
                 return packetwait;
         }
-        return max_wait * 1000000;
+
+        // TODO: check for overflows. Though what's the worst that
+        // could happen?
+        return (uint32_t)(max_wait * 1000000);
 }
 
 /**
@@ -1200,15 +1347,13 @@ pingmac_send(uint16_t id, uint16_t seq)
                        payload_suffix, payload_suffix_size);
         }
 
-	int c;
-
 	if (-1 == (icmp = libnet_build_icmpv4_echo(ICMP_ECHO, /* type */
 						   0, /* code */
 						   0, /* checksum */
 						   id, /* id */
 						   seq, /* seq */
 						   (uint8_t*)padding, /* payload */
-						   sizeof padding, /* payload len */
+						   cast_size_uint32(padding_size, NULL), /* payload len */
 						   libnet,
 						   icmp))) {
 		fprintf(stderr, "libnet_build_icmpv4_echo(): %s\n",
@@ -1218,7 +1363,7 @@ pingmac_send(uint16_t id, uint16_t seq)
 
 	if (-1==(ipv4 = libnet_build_ipv4(LIBNET_IPV4_H
                                           + LIBNET_ICMPV4_ECHO_H
-                                          + sizeof padding,
+                                          + cast_size_uint16(padding_size, NULL),
 					  0, /* ToS */
 					  id, /* id */
 					  0, /* frag */
@@ -1239,9 +1384,9 @@ pingmac_send(uint16_t id, uint16_t seq)
                 eth = libnet_build_802_1q(dstmac,
                                           srcmac,
                                           ETHERTYPE_VLAN,
-                                          vlan_prio,
+                                          cast_int16_uint8(vlan_prio, NULL),
                                           0, // cfi
-                                          vlan_tag,
+                                          cast_int16_uint16(vlan_tag, NULL),
                                           ETHERTYPE_IP,
                                           NULL, // payload
                                           0, // payload length
@@ -1269,7 +1414,7 @@ pingmac_send(uint16_t id, uint16_t seq)
                        (long)lastpacketsent.tv_sec,
                        (long)lastpacketsent.tv_nsec);
 	}
-	if (-1 == (c = libnet_write(libnet))) {
+        if (-1 == libnet_write(libnet)) {
 		fprintf(stderr, "arping: libnet_write(): %s\n",
 			libnet_geterror(libnet));
 		sigint(0);
@@ -1314,9 +1459,9 @@ pingip_send()
                 eth = libnet_build_802_1q(dstmac,
                                           srcmac,
                                           ETHERTYPE_VLAN,
-                                          vlan_prio,
+                                          cast_int16_uint8(vlan_prio, NULL),
                                           0, // cfi
-                                          vlan_tag,
+                                          cast_int16_uint16(vlan_tag, NULL),
                                           ETHERTYPE_ARP,
                                           NULL, // payload
                                           0, // payload size
@@ -1367,6 +1512,7 @@ pingip_recv(const char *unused, struct pcap_pkthdr *h, const char * const packet
 	struct libnet_arp_hdr *harp;
         struct timespec arrival;
         UNUSED(unused);
+        assert(packet);
 
         if (verbose > 2) {
 		printf("arping: received response for IP ping\n");
@@ -1574,6 +1720,7 @@ pingmac_recv(const char* unused, struct pcap_pkthdr *h, uint8_t *packet)
 	struct libnet_icmpv4_hdr *hicmp;
         struct timespec arrival;
         UNUSED(unused);
+        assert(packet);
 
 	if(verbose>2) {
 		printf("arping: received response for mac ping\n");
@@ -1674,10 +1821,11 @@ pingmac_recv(const char* unused, struct pcap_pkthdr *h, uint8_t *packet)
         }
 
         const char* payload = (char*)hicmp + LIBNET_ICMPV4_ECHO_H;
-        const ssize_t payload_size = h->len - (payload - (char*)packet);
-        if (payload_size < 0) {
+        const size_t tmp = cast_ssize_size(payload - (char*)packet, NULL);
+        if (h->len < tmp) {
                 return;
         }
+        const size_t payload_size = h->len - tmp;
         if (payload_size < sizeof(struct timespec) + payload_suffix_size) {
                 return;
         }
@@ -1686,7 +1834,7 @@ pingmac_recv(const char* unused, struct pcap_pkthdr *h, uint8_t *packet)
                        payload_size);
         }
         if (memcmp(&payload[sizeof(struct timespec)],
-                    payload_suffix, payload_suffix_size)) {
+                   payload_suffix, payload_suffix_size)) {
                     return;
         }
         if (verbose > 3) {
@@ -1707,7 +1855,7 @@ pingmac_recv(const char* unused, struct pcap_pkthdr *h, uint8_t *packet)
                 break;
         case NORMAL:
                 printf("%d bytes from %s (%s): icmp_seq=%d time=%s", h->len,
-                       libnet_addr2name4(*(int*)&hip->ip_src, 0),
+                       libnet_addr2name4(*(uint32_t*)&hip->ip_src, 0),
                        format_mac(pkt_srcmac, buf, sizeof(buf)),
                        htons(hicmp->icmp_seq),
                        ts2str(&lastpacketsent, &arrival, buf2, sizeof(buf2)));
@@ -1894,7 +2042,7 @@ ping_recv(pcap_t *pcap, uint32_t packetwait, pcap_handler func)
 
 // return 1 on success.
 static int
-xresolve(libnet_t* l, const char *name, int r, uint32_t *addr)
+xresolve(libnet_t* l, const char *name, uint8_t r, uint32_t *addr)
 {
         if (!strcmp(ip_broadcast, name)) {
                 *addr = 0xffffffff;
@@ -1924,9 +2072,9 @@ arping_main(int argc, char **argv)
         int opt_B = 0;
         int opt_T = 0;
         int opt_U = 0;
+        int c;
         const char* drop_group = NULL;  // -g
         const char *parm; // First argument, meaning the target IP.
-	int c;
 	int maxcount = -1;
 	int dont_use_arping_lookupdev=0;
 	struct bpf_program bp;
@@ -1936,13 +2084,16 @@ arping_main(int argc, char **argv)
         double deadline = -1;
         char bpf_filter[64];
         ebuf[0] = 0;
-        srandom(time(NULL));
+        srandom((unsigned)time(NULL));
 
-        for (c = 1; c < argc; c++) {
-                if (!strcmp(argv[c], "--help")) {
-                        standard_usage();
-                        extended_usage();
-                        exit(0);
+        {
+                int c;
+                for (c = 1; c < argc; c++) {
+                        if (!strcmp(argv[c], "--help")) {
+                                standard_usage();
+                                extended_usage();
+                                exit(0);
+                        }
                 }
         }
 
@@ -1971,10 +2122,17 @@ arping_main(int argc, char **argv)
                         opt_B = 1;
 			break;
 		case 'c':
-			maxcount = atoi(optarg);
+                        maxcount = must_parse_int(optarg, "count (-c)");
+                        if (maxcount < 0) {
+                                fprintf(stderr,
+                                        "arping: count (-c) must be >0. Got %d\n",
+                                        maxcount);
+                                exit(1);
+                        }
 			break;
                 case 'C':
-                        max_replies = atoi(optarg);
+                        max_replies = must_parse_uint(optarg,
+                                                      "max replies (-C)");
                         break;
 		case 'd':
 			finddup = 1;
@@ -2022,7 +2180,7 @@ arping_main(int argc, char **argv)
 			display = QUIET;
 			break;
                 case 'Q':
-                        vlan_prio = atoi(optarg);
+                        vlan_prio = must_parse_int16(optarg, "802.1p prio (-Q)");
                         if (vlan_prio < 0 || vlan_prio > 7) {
                                 fprintf(stderr,
                                         "arping: 802.1p priority must be 0-7. It's %d\n",
@@ -2065,20 +2223,37 @@ arping_main(int argc, char **argv)
 			verbose++;
 			break;
 		case 'V':
-			vlan_tag = atoi(optarg);
+			vlan_tag = must_parse_int16(optarg, "VLAN (-V)");
                         if (vlan_tag < 0 || vlan_tag > 4095) {
                                 fprintf(stderr,
-                                        "arping: vlan tag must 0-4095. Is %d\n",
+                                        "arping: vlan tag must be 0-4095. Is %d\n",
                                         vlan_tag);
                                 exit(1);
                         }
                         break;
 		case 'w':
-                        deadline = atof(optarg);
+                        deadline = must_parse_float(optarg, "deadline (-w)");
+                        if (deadline < 0) {
+                                fprintf(stderr,
+                                        "arping: deadline (-w) must be >=0. Is %f\n",
+                                        deadline);
+                                exit(1);
+                        }
 			break;
-                case 'W':
-                        packetwait = (unsigned)(1000000.0 * atof(optarg));
+                case 'W': {
+                        const float val = must_parse_float(optarg,
+                                                           "packetwait (-W)");
+                        // TODO: range check into unsigned.
+                        // Also turns out zero doesn't work.
+                        packetwait = (unsigned)(1000000.0 * val);
+                        if (val < 0) {
+                                fprintf(stderr,
+                                        "arping: packetwait (-W) must be >=0. Is %f\n",
+                                        val);
+                                exit(1);
+                        }
                         break;
+                }
                 case 'z':
                         use_seccomp = 1;
                         break;
@@ -2103,39 +2278,39 @@ arping_main(int argc, char **argv)
 
         // Generate random payload suffix for MAC pings, to be able to
         // differentiate from unrelated ping replies.
-        if (payload_suffix_size < 0) {
-                payload_suffix_size = 4;
-                payload_suffix = malloc(payload_suffix_size);
-                if (payload_suffix) {
-                        const ssize_t rc = xgetrandom(payload_suffix, payload_suffix_size, 0);
-                        if (rc == -1) {
-                                fprintf(stderr,
-                                        "arping: failed to get %zd random bytes: %s\n",
-                                        payload_suffix_size,
-                                        strerror(errno));
-                                free(payload_suffix);
-                                payload_suffix = NULL;
-                        } else if (payload_suffix_size != rc) {
-                                fprintf(stderr,
-                                        "arping: only got %zd out of %zd bytes for random suffix\n",
-                                        rc, payload_suffix_size);
-                        }
-                } else {
-                        fprintf(stderr, "arping: failed to allocate %zd bytes for payload suffix.\n",
-                                payload_suffix_size);
-                }
+        payload_suffix = malloc(payload_suffix_size);
+        if (!payload_suffix) {
+                fprintf(stderr, "arping: failed to allocate %zu bytes for payload suffix.\n",
+                        payload_suffix_size);
+                exit(1);
+        }
 
-                if (!payload_suffix) {
-                        fprintf(stderr, "arping:  Using constant suffix.\n");
-                        payload_suffix = "arping";
-                        payload_suffix_size = strlen(payload_suffix);
+        // Create some default content.
+        {
+                const char* def = "arping/";
+                for (size_t c = 0; c < payload_suffix_size; c++) {
+                        payload_suffix[c] = def[c % strlen(def)];
                 }
         }
 
+        // Randomize should be even better.
+        const ssize_t rc = xgetrandom(payload_suffix, payload_suffix_size, 0);
+        if (rc < -1) {
+                fprintf(stderr,
+                        "arping: failed to get %zu random bytes: %s\n",
+                        payload_suffix_size,
+                        strerror(errno));
+                free(payload_suffix);
+        } else if (payload_suffix_size != cast_ssize_size(rc, NULL)) {
+                fprintf(stderr,
+                        "arping: only got %zd out of %zu bytes for random suffix\n",
+                        rc, payload_suffix_size);
+        }
+
         if (((mode == PINGIP) && opt_T)
             || ((mode == PINGMAC) && (opt_B || dstmac_opt || opt_U))) {
                 fprintf(stderr, "arping: -T can only be used to ping MAC"
-                        " and -BtU only to ping IPs");
+                        " and -BtU only to ping IPs\n");
                 exit(1);
         }
         if (opt_T && opt_B) {
@@ -2356,7 +2531,7 @@ arping_main(int argc, char **argv)
 	/*
 	 * pcap init
 	 */
-        if (!(pcap = do_pcap_open_live(ifname, 100, promisc, 10, ebuf))) {
+        if (!(pcap = do_pcap_open_live(ifname, 100, 10, ebuf))) {
                 strip_newline(ebuf);
                 fprintf(stderr, "arping: pcap_open_live(): %s\n", ebuf);
 		exit(1);
@@ -2398,7 +2573,8 @@ arping_main(int argc, char **argv)
                 } else {
                         snprintf(bpf_filter, sizeof(bpf_filter), "arp");
                 }
-                if (-1 == pcap_compile(pcap, &bp, bpf_filter, 0, -1)) {
+                if (-1 == pcap_compile(pcap, &bp, bpf_filter, 0,
+                                       PCAP_NETMASK_UNKNOWN)) {
                         fprintf(stderr, "arping: pcap_compile(%s): %s\n",
                                 bpf_filter, pcap_geterr(pcap));
 			exit(1);
@@ -2411,7 +2587,8 @@ arping_main(int argc, char **argv)
                 } else {
                         snprintf(bpf_filter, sizeof(bpf_filter), "icmp");
                 }
-                if (-1 == pcap_compile(pcap, &bp, bpf_filter, 0,-1)) {
+                if (-1 == pcap_compile(pcap, &bp, bpf_filter, 0,
+                                       PCAP_NETMASK_UNKNOWN)) {
                         fprintf(stderr, "arping: pcap_compile(%s): %s\n",
                                 bpf_filter, pcap_geterr(pcap));
 			exit(1);
@@ -2482,7 +2659,8 @@ arping_main(int argc, char **argv)
 	} else { /* PINGMAC */
 		int c;
 		for (c = 0; (maxcount < 0 || c < maxcount) && !time_to_die; c++) {
-			pingmac_send(xrandom(), c);
+                        pingmac_send(cast_int_uint16(xrandom() & 0xffff, NULL),
+                                     cast_int_uint16(c & 0xffff, NULL));
                         const uint32_t w = wait_time(deadline, packetwait);
                         if (w == 0) {
                                 break;
@@ -2491,12 +2669,12 @@ arping_main(int argc, char **argv)
 		}
 	}
         if (display == DOT) {
-                const float succ = 100.0 - 100.0 * (float)(numrecvd)/(float)numsent;
+                const double succ = 100.0 - 100.0 * (double)(numrecvd)/(double)numsent;
                 printf("\t%3.0f%% packet loss (%d extra)\n",
                        (succ < 0.0) ? 0.0 : succ,
                        (succ < 0.0) ? (numrecvd - numsent) : 0);
         } else if (display == NORMAL) {
-                const float succ = 100.0 - 100.0 * (float)(numrecvd)/(float)numsent;
+                const double succ = 100.0 - 100.0 * (double)(numrecvd)/(double)numsent;
                 printf("\n--- %s statistics ---\n"
                        "%d packets transmitted, "
                        "%d packets received, "
@@ -2518,11 +2696,42 @@ arping_main(int argc, char **argv)
                 printf("\n");
 	}
 
+        if (fflush(stdout)) {
+                fprintf(stderr, "arping: Failed to write to stdout: %s\n",
+                        strerror(errno));
+                return EXIT_FAILURE;
+        }
+
+        if (ferror(stdout)) {
+                fprintf(stderr, "arping: Write failure to stdout: %s\n",
+                        strerror(errno));
+                return EXIT_FAILURE;
+        }
+
+        if (fflush(stderr)) {
+                // Likely this will go nowhere. But we can't write to
+                // stdout in case it's being machine parsed.
+                //
+                // At least it should show up in strace.
+                fprintf(stderr, "arping: Failed to write to stderr: %s\n",
+                        strerror(errno));
+                return EXIT_FAILURE;
+        }
+
+        if (ferror(stderr)) {
+                // Same here. Likely will go nowhere.
+                fprintf(stderr, "arping: Write failure to stderr: %s\n",
+                        strerror(errno));
+                return EXIT_FAILURE;
+        }
+
+        // We could close stdout/stdin too, but that'll interfere with
+        // unit testing.
+
         if (finddup) {
                 return dupfound;
-        } else {
-                return !numrecvd;
         }
+        return !numrecvd;
 }
 /* ---- Emacs Variables ----
  * Local Variables:
