@@ -79,10 +79,8 @@ read_main(void* p)
         char *cur = out->buffer;
 
         for (;;) {
-                ssize_t n;
-                n = out->bufsize - (cur - out->buffer);
-                assert(n > 0);
-                n = read(out->reader_fd, cur, n);
+                const size_t to_read = out->bufsize - (size_t)(cur - out->buffer);
+                const ssize_t n = read(out->reader_fd, cur, to_read);
                 if (n > 0) {
                         cur += n;
                 }
@@ -199,9 +197,9 @@ mkpacket(struct pcap_pkthdr* pkthdr)
 }
 
 static void
-dump_packet(uint8_t* packet, int len)
+dump_packet(uint8_t* packet, uint32_t len)
 {
-        int c;
+        uint32_t c;
         for (c = 0; c < len; c++) {
                 fprintf(stderr, "0x%.2x, ", (int)packet[c]);
                 if (!((c+1) % 10)) {
@@ -232,6 +230,7 @@ START_TEST(test_mkpacket)
                 dump_packet(packet, pkthdr.caplen);
         }
         fail_unless(!memcmp(packet, correct_packet, pkthdr.caplen));
+        free(packet);
 } END_TEST
 
 
@@ -240,16 +239,15 @@ START_TEST(pingip_uninteresting_packet)
 {
         struct pcap_pkthdr pkthdr;
         uint8_t* packet;
-        struct libnet_802_3_hdr* heth;
         struct libnet_arp_hdr* harp;
 
-        int prev_numrecvd = numrecvd;
+        unsigned int prev_numrecvd = numrecvd;
         struct captured_output* sout;
 
         // Completely broken packet.
         packet = calloc(1, 1500);
-        sout = capture(1);
-        pingip_recv(NULL, &pkthdr, packet);
+        sout = capture(STDOUT_FILENO);
+        pingip_recv(NULL, &pkthdr, (char*)packet);
         stop_capture(sout);
         fail_unless(strlen(sout->buffer) == 0);
         fail_unless(prev_numrecvd == numrecvd);
@@ -260,8 +258,8 @@ START_TEST(pingip_uninteresting_packet)
         packet = mkpacket(&pkthdr);
         harp = (void*)((char*)packet + LIBNET_ETH_H);
         harp->ar_pro = 0;
-        sout = capture(1);
-        pingip_recv(NULL, &pkthdr, packet);
+        sout = capture(STDOUT_FILENO);
+        pingip_recv(NULL, &pkthdr, (char*)packet);
         stop_capture(sout);
         fail_unless(prev_numrecvd == numrecvd);
         fail_unless(strlen(sout->buffer) == 0);
@@ -272,8 +270,8 @@ START_TEST(pingip_uninteresting_packet)
         packet = mkpacket(&pkthdr);
         harp = (void*)((char*)packet + LIBNET_ETH_H);
         harp->ar_hrd = 0;
-        sout = capture(1);
-        pingip_recv(NULL, &pkthdr, packet);
+        sout = capture(STDOUT_FILENO);
+        pingip_recv(NULL, &pkthdr, (char*)packet);
         stop_capture(sout);
         fail_unless(prev_numrecvd == numrecvd);
         fail_unless(strlen(sout->buffer) == 0);
@@ -286,8 +284,8 @@ START_TEST(pingip_uninteresting_packet)
                 packet = mkpacket(&pkthdr);
                 harp = (void*)((char*)packet + LIBNET_ETH_H);
                 memcpy((char*)harp + harp->ar_hln + LIBNET_ARP_H, &wrongip, 4);
-                sout = capture(1);
-                pingip_recv(NULL, &pkthdr, packet);
+                sout = capture(STDOUT_FILENO);
+                pingip_recv(NULL, &pkthdr, (char*)packet);
                 stop_capture(sout);
                 fail_unless(prev_numrecvd == numrecvd);
                 fail_unless(strlen(sout->buffer) == 0);
@@ -298,8 +296,8 @@ START_TEST(pingip_uninteresting_packet)
         // Short packet.
         packet = mkpacket(&pkthdr);
         pkthdr.caplen = pkthdr.len = 41;
-        sout = capture(1);
-        pingip_recv(NULL, &pkthdr, packet);
+        sout = capture(STDOUT_FILENO);
+        pingip_recv(NULL, &pkthdr, (char*)packet);
         stop_capture(sout);
         fail_unless(prev_numrecvd == numrecvd);
         fail_unless(strlen(sout->buffer) == 0);
@@ -309,8 +307,8 @@ START_TEST(pingip_uninteresting_packet)
         // Short captured packet.
         packet = mkpacket(&pkthdr);
         pkthdr.caplen = 41;
-        sout = capture(1);
-        pingip_recv(NULL, &pkthdr, packet);
+        sout = capture(STDOUT_FILENO);
+        pingip_recv(NULL, &pkthdr, (char*)packet);
         stop_capture(sout);
         fail_unless(prev_numrecvd == numrecvd);
         fail_unless(strlen(sout->buffer) == 0);
@@ -343,8 +341,8 @@ START_TEST(pingip_uninteresting_packet)
         };
                 pkthdr.len = 60;
                 pkthdr.caplen = 60;
-                sout = capture(1);
-                pingip_recv(NULL, &pkthdr, packet);
+                sout = capture(STDOUT_FILENO);
+                pingip_recv(NULL, &pkthdr, (char*)packet);
                 stop_capture(sout);
                 fail_unless(strlen(sout->buffer) == 0, sout->buffer);
                 fail_unless(prev_numrecvd == numrecvd);
@@ -354,8 +352,8 @@ START_TEST(pingip_uninteresting_packet)
         // Wrong length of protocol address.
         packet = mkpacket(&pkthdr);
         ((struct libnet_arp_hdr*)((char*)packet + LIBNET_ETH_H))->ar_pln = 6;
-        sout = capture(1);
-        pingip_recv(NULL, &pkthdr, packet);
+        sout = capture(STDOUT_FILENO);
+        pingip_recv(NULL, &pkthdr, (char*)packet);
         stop_capture(sout);
         fail_unless(prev_numrecvd == numrecvd);
         fail_unless(strlen(sout->buffer) == 0);
@@ -389,7 +387,7 @@ START_TEST(pingip_interesting_packet)
                 0x6f, 0xa8, 0x58, 0x63,
         };
         numrecvd = 0;
-        int prev_numrecvd = numrecvd;
+        unsigned int prev_numrecvd = numrecvd;
 
         dstip = htonl(0x12345678);
 
@@ -404,8 +402,8 @@ START_TEST(pingip_interesting_packet)
         const char* correct0 =
                         "60 bytes from 77:88:99:aa:bb:cc (18.52.86.120): "
                         "index=0 time=";
-        sout = capture(1);
-        pingip_recv(NULL, &pkthdr, packet);
+        sout = capture(STDOUT_FILENO);
+        pingip_recv(NULL, &pkthdr, (char*)packet);
         stop_capture(sout);
 
         char* emsg = NULL;
@@ -416,12 +414,12 @@ START_TEST(pingip_interesting_packet)
         uncapture(sout);
         free(emsg);
 
-        fail_unless(numrecvd == prev_numrecvd + 1,
-                    "numrecvd not incremented");
+        // Check numrecvd incremented.
+        ck_assert_int_eq(numrecvd, prev_numrecvd + 1);
 
-        pingip_recv(NULL, &pkthdr, packet);
-        fail_unless(numrecvd == prev_numrecvd + 2,
-                    "numrecvd not incremented second time");
+        pingip_recv(NULL, &pkthdr, (char*)packet);
+        // Check that numrecvd is incremented second time.
+        ck_assert_int_eq(numrecvd,prev_numrecvd + 2);
 } END_TEST
 
 START_TEST(strip_newline_test)
@@ -438,7 +436,7 @@ START_TEST(strip_newline_test)
         int c;
         for (c = 0; tests[c][0]; c++){
                 char buf[128];
-                strcpy(buf, tests[c][0]);
+                snprintf(buf, sizeof(buf), "%s", tests[c][0]);
                 strip_newline(buf);
                 fail_unless(!strcmp(buf, tests[c][1]));
         }
@@ -492,6 +490,114 @@ START_TEST(get_mac_addr_fail)
         }
 } END_TEST
 
+START_TEST(badarg_maxcount_neg)
+{
+        char* args[] = {
+                "arping",
+                "-c", "-1",
+                "-h", // To exit cleanly.
+                "dummy",
+                NULL
+        };
+        exit(arping_main(sizeof(args)/sizeof(char*)-1, args));
+}
+
+START_TEST(badarg_maxcount_nan)
+{
+        char* args[] = {
+                "arping",
+                "-c", "number",
+                "-h", // To exit cleanly.
+                "dummy",
+                NULL
+        };
+        exit(arping_main(sizeof(args)/sizeof(char*)-1, args));
+}
+
+START_TEST(badarg_maxcount_range)
+{
+        char* args[] = {
+                "arping",
+                "-c", "100000000000000000000000000000",
+                "-h", // To exit cleanly.
+                "dummy",
+                NULL
+        };
+        exit(arping_main(sizeof(args)/sizeof(char*)-1, args));
+}
+
+START_TEST(arg_maxcount_good)
+{
+        char* args[] = {
+                "arping",
+                "-c", "1",
+                "-h", // To exit cleanly.
+                "dummy",
+                NULL
+        };
+        exit(arping_main(sizeof(args)/sizeof(char*)-1, args));
+}
+
+START_TEST(arg_maxcount_hex)
+{
+        char* args[] = {
+                "arping",
+                "-c", "0xa",
+                "-h", // To exit cleanly.
+                "dummy",
+                NULL
+        };
+        exit(arping_main(sizeof(args)/sizeof(char*)-1, args));
+}
+
+START_TEST(badarg_packetwait_neg)
+{
+        char* args[] = {
+                "arping",
+                "-W", "-1",
+                "-h", // To exit cleanly.
+                "dummy",
+                NULL
+        };
+        exit(arping_main(sizeof(args)/sizeof(char*)-1, args));
+}
+
+START_TEST(badarg_packetwait_nan)
+{
+        char* args[] = {
+                "arping",
+                "-W", "huteosu",
+                "-h", // To exit cleanly.
+                "dummy",
+                NULL
+        };
+        exit(arping_main(sizeof(args)/sizeof(char*)-1, args));
+}
+
+START_TEST(badarg_packetwait_range)
+{
+        char* args[] = {
+                "arping",
+                "-W", "10000000000000000000000000000000000000000",
+                "-h", // To exit cleanly.
+                "dummy",
+                NULL
+        };
+        exit(arping_main(sizeof(args)/sizeof(char*)-1, args));
+}
+
+START_TEST(arg_packetwait_good)
+{
+        char* args[] = {
+                "arping",
+                "-W", "1.2",
+                "-h", // To exit cleanly.
+                "dummy",
+                NULL
+        };
+        exit(arping_main(sizeof(args)/sizeof(char*)-1, args));
+}
+
 START_TEST(libnet_init_bad_nolo)
 {
         // It'll only try lo if named interface fails.
@@ -526,7 +632,6 @@ END_TEST
 static Suite*
 arping_suite(void)
 {
-        int c;
         Suite* s = suite_create("Arping");
 
 
@@ -547,13 +652,21 @@ arping_suite(void)
         SIGH_LIBCHECK(get_mac_addr_success);
         SIGH_LIBCHECK(get_mac_addr_fail);
         SIGH_LIBCHECK(libnet_init_good);
-
+        SIGH_LIBCHECK(arg_maxcount_good);
+        SIGH_LIBCHECK(arg_maxcount_hex);
+        SIGH_LIBCHECK(arg_packetwait_good);
 
 #define SIGH_LIBCHECK_EXIT(tn)          \
         tc_core = tcase_create(#tn);   \
         tcase_add_exit_test(tc_core, tn, 1);      \
         suite_add_tcase(s, tc_core);
 
+        SIGH_LIBCHECK_EXIT(badarg_maxcount_neg);
+        SIGH_LIBCHECK_EXIT(badarg_maxcount_nan);
+        SIGH_LIBCHECK_EXIT(badarg_maxcount_range);
+        SIGH_LIBCHECK_EXIT(badarg_packetwait_neg);
+        SIGH_LIBCHECK_EXIT(badarg_packetwait_nan);
+        SIGH_LIBCHECK_EXIT(badarg_packetwait_range);
         SIGH_LIBCHECK_EXIT(libnet_init_bad_nolo);
         SIGH_LIBCHECK_EXIT(libnet_init_null_nolo_nonull);
 
@@ -563,14 +676,13 @@ arping_suite(void)
 int
 main()
 {
-        int number_failed;
         Suite *s = arping_suite();
-        SRunner *sr = srunner_create (s);
-        srunner_run_all (sr, CK_NORMAL);
-        number_failed = srunner_ntests_failed (sr);
-        srunner_free (sr);
+        SRunner *sr = srunner_create(s);
+        srunner_run_all(sr, CK_NORMAL);
+        const int number_failed = srunner_ntests_failed(sr);
+        srunner_free(sr);
+        fprintf(stderr, "Number of failed tests: %d\n", number_failed);
         return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
-        return 0;
 }
 /* ---- Emacs Variables ----
  * Local Variables:
